"""
$RCSfile: CVSFile.py,v $

The CVSFile package is a Zope product.  CVSFile allows one to use
CVS as a version control system for Zope content, by pointing to a
file in the developer's CVS sandbox on the Zope server.

Author: <a href="mailto:cstrong@arielpartners.com">Craeg Strong</a>
Release: 1.0
"""

__cvstag__  = '$Name:  $'[6:-2]
__date__    = '$Date: 2002/08/27 23:16:46 $'[6:-2]
__version__ = '$Revision: 1.58 $'[10:-2]

# Base classes, peer classes
from Products.ExternalFile.ExternalFile import ExternalFile
from ICVSFile import ICVSFile
from CVSSandboxRegistry import CVSSandboxRegistry, findCVSSandboxRegistry
from SandboxInfo import SandboxInfo
from OSUtils import runCommand, command_separator
from Products.ExternalFile.FileUtils import copy_file

# Zope builtins
from Globals import DTMLFile, MessageDialog
from AccessControl import ClassSecurityInfo
import Globals  # InitializeClass (security stuff)

# Python builtins
import os

################################################################
# Container Methods
################################################################

# globals() is standard python for passing the current namespace
manage_addForm       = DTMLFile('dtml/create',               globals())
cvsregistry_formpart = DTMLFile('dtml/cvsregistry_formpart', globals())
cvssandbox_formpart  = DTMLFile('dtml/cvssandbox_formpart',  globals())

def manage_add(self,
               id, title, description, 
               target_filepath, basedir='', REQUEST=None):
    """
    
    Factory method to actually create an instance of CVSFile.  Assumes
    that all necessary checks and confirmations have been done.
    Assumes that target_filepath points to a valid, existing file with
    appropriate permissions in the filesystem/CVS sandbox in question.

    You should call this method directly if you are creating an
    instance of CVSFile programmatically and have 'vetted' all of your
    parameters for correctness.
    
    """
    self._setObject(id, CVSFile(id, title, description,
                                target_filepath))
    self._getOb(id).reindex_object()

def manage_add_with_upload(self,
                           id, title, description,
                           target_filepath, upload_file, basedir,
                           REQUEST=None):
    """

    Factory method to actually create an instance of CVSFile.  This
    method includes an upload_file parameter.  This method assumes all
    parameters are correct (it does no error checking).

    You should call this method directly if you are creating an
    instance of CVSFile programmatically via an uploaded file and have
    'vetted' all of your parameters for correctness.

    """
    fully_resolved_target_filepath = os.path.join(basedir,target_filepath)    

    copy_file(upload_file, fully_resolved_target_filepath)
    
    self._setObject(id, CVSFile(id, title, description,
                                target_filepath))
    self._getOb(id).reindex_object()

################################################################
# CVSFile class
################################################################

class CVSFile(ExternalFile):

    """
    Extended from ExternalFile
    """

    meta_type = 'CVS File'  # This is the name Zope will use for the Product in
                            # the "addProduct" list

    __implements__ = ICVSFile

    # This tuple defines a dictionary for each tab in the management interface
    # label = label of tab, action = url it links to
    manage_options = ExternalFile.manage_options + (
	{'label':'CVS',	    'action': 'manage_cvsForm',     'help': ('CVSFile', 'cvs.stx')     },
        )

    _security = ClassSecurityInfo()

    # set security for the object itself, e.g. if it is accessed in DTML code
    # This line is REQUIRED to allow access to CVSFiles by the public (ariel DTDs, etc.)
    _security.declareObjectPublic()

    # call __call__() method on DTMLFile for our edit form, point at our cvs_form.dtml
    # globals() is standard python for passing the current namespace
    cvs_form           = DTMLFile('dtml/cvs_form',globals())
    
    _security.declareProtected('Manage CVS Files',      'manage_cvsForm')
    manage_cvsForm     = DTMLFile('dtml/cvs',globals())

    # these lines exist so we can call them from DTML
    cvsregistry_formpart = cvsregistry_formpart
    cvssandbox_formpart  = cvssandbox_formpart
    # end of DTML scoping hack

    def __init__(self, id, title, description, relativeFilepath):
        """
        Initialize a new instance of CVSFile
        """
        ExternalFile.__init__(self, id, title,
                              description,
                              relativeFilepath)

        # Default settings for arguments to CVS commands
        if self.getDisplayType() == 'ascii':
            self.add_args = ''
        else:
            self.add_args = '-kb'
        self.update_args = ''
        self.commit_args = ''
        
    ################################################################
    # Methods implementing the ICVSFile interface below
    ################################################################

    _security.declareProtected('View',  'findRegistry')
    def findRegistry(self):
        """

        Find an instance of CVSSandboxRegistry to obtain sandbox
        information.  This calls a module scoped function which does
        the real work of finding the sandbox.  Because we need this
        behavior to even _create_ an instance of CVSFile the function
        needs to be module scoped.

        """
        
        return findCVSSandboxRegistry(self)

    _security.declareProtected('View',  'getCurrentSandbox')
    def getCurrentSandbox(self, REQUEST):
        """

        Convenience function for getting the current sandbox from the
        sandbox registry

        """
        registry = self.findRegistry()
        if not registry:
            raise EnvironmentError, 'CVS Sandbox Registry not found'
        else:
            return registry.getCurrentSandbox(REQUEST)    

    def addFile(self, si, filepath, addArgs=''):
        """

        Perform a 'cvs add' on the CVS file pointed to by this
        object, using the sandbox information from the SandboxInfo
        parameter.

        """
        #
        # According to the cvs manual, you must physically "be" inside the
        # directory to do a cvs add, you cannot traverse down directories.
        # That is:
        #
        # BAD:  cd /home/cstrong/sandbox; cvs add foo/bar/mumble.txt
        #
        # GOOD: cd /home/cstrong/sandbox/foo/bar; cvs add mumble.txt
        #
        # See docs: http://www.cvshome.org/docs/manual/cvs_7.html
        # CKS 2/18/2002
        #
        dir      = os.path.dirname(filepath)
        filename = os.path.basename(filepath)
        
        command  = "cd %s %s cvs -d %s add %s %s" % ( dir, command_separator,
                                                      si.cvsroot, addArgs,
                                                      filename )
        return runCommand(command)

    def updateFile(self, si, updateArgs = ''):
        """

        Perform a 'cvs update' on the CVS file pointed to by this
        object, using the sandbox information from the SandboxInfo
        parameter.

        """
        command = "cd %s %s cvs -d %s -q update %s %s" % ( si.basedir,
                                                           command_separator,
                                                           si.cvsroot,
                                                           updateArgs,
                                                           self.filepath )
        return runCommand(command)

    def revertFile(self, filepath, si):
        """

        Remove the sandbox version of the file pointed to by this
        object, the do a 'cvs update' and pass the output back to the
        form.  This is used to throw away any non-checked in changes
        to the file.  Retrieve sandbox information from the passed-in
        SandboxInfo parameter.

        """

        # remove sandbox version of file
        os.remove(filepath)

        # then do a cvs update
        return self.updateFile(si, '')

    def commitFile(self, si, commitMessage, userName, commitArgs=''):
        """

        Perform a commit on the CVS file pointed to by this object.
        This will submit local changes to the central repository.
        Retrieve sandbox information from the passed-in SandboxInfo
        parameter.  Include the passed in username in the commit
        message, because otherwise there would be no way to
        distinguish whodunit, because all userids are the same, namely
        the userid under which the Zope server was started.
        
        """
        who = 'Zope CVSFile user: %s\n' % userName

        command = "cd  %s %s cvs -d %s commit -m \"%s\" %s %s" % ( si.basedir,
                                                                   command_separator,
                                                                   si.cvsroot,
                                                                   who + commitMessage,
                                                                   commitArgs,
                                                                   self.filepath )
        return runCommand(command)

    ################################################################
    # Methods implementing the IExternalFile interface
    ################################################################

    _security.declareProtected('View',  'getFilepath')
    def getFilepath(self, REQUEST):
        """
        
        Get the full path of the external file this CVSFile is
        pointing to.  Depends on the sandbox we are currently using.
        
        """
        si = self.getCurrentSandbox(REQUEST)
        return os.path.join(si.basedir, self.filepath)

    ################################################################
    # Methods called from the CVS form 
    ################################################################

    # called from cvs.dtml form
    _security.declareProtected('Manage CVS Files',      'cvsUpdateFile')
    def manage_AddFile(self, addArgs='', REQUEST=None):
        """
        Does a CVS add on the file and passes output back to the form.
        """
        si       = self.getCurrentSandbox(REQUEST)
        filepath = self.getFilepath(REQUEST)
        output   = self.addFile(si, filepath, addArgs)
        return self.manage_cvsForm(addFileOutput = output)    

    # called from cvs.dtml form
    _security.declareProtected('Manage CVS Files',      'cvsUpdateFile')
    def manage_UpdateFile(self, updateArgs='', REQUEST=None):
        """
        Does a CVS update on the file and passes output back to the form.
        """
        si     = self.getCurrentSandbox(REQUEST)
        output = self.updateFile(si, updateArgs)
        return self.manage_cvsForm(updateFileOutput = output)    


    # called from cvs.dtml form
    _security.declareProtected('Manage CVS Files',      'cvsRevertFile')
    def manage_RevertFile(self, REQUEST):
        """
        
        Removes the sandbox version of a file, then does a CVS update
        on the file and passes output back to the form.  This, in
        effect, throws away any non-checked in changes to the file.

        """
        si       = self.getCurrentSandbox(REQUEST)
        filepath = self.getFilepath(REQUEST)
        output   = self.revertFile(filepath, si)
        return self.manage_cvsForm(revertFileOutput = output)

    # called from cvs.dtml form
    _security.declareProtected('Manage CVS Files',  'cvsCommitFile')
    def manage_CommitFile(self, commitMessage='', userName='', commitArgs='',
                          REQUEST=None):
        """
        Does a CVS commit on the file and passes output back to the form
        """
        #
        # Since Zope always runs as an administrative user
        # (e.g. "nobody") we won't know who actually committed these
        # changes unless we capture it another way.  Since by
        # convention we make the Zope usernames the same as the
        # UNIX/CVS usernames, we grab it from Zope via a hidden field
        # in the cvs.dtml form.
        #
        if not commitMessage:
            return MessageDialog(title   = 'ERROR',
                                 message = 'You must specify a commit message.',
                                 action  = 'manage_cvsForm')
        
        si     = self.getCurrentSandbox(REQUEST)
        output = self.commitFile(si, commitMessage, userName, commitArgs)
        return self.manage_cvsForm(commitFileOutput = output)

    ################################################################
    # Support for Schema Migration
    ################################################################

    def repair(self):
        """

        Repair this object.  This method is used for schema migration,
        when the class definition changes and existing instances of
        previous versions must be updated.

        """
        if not hasattr(self, 'add_args'):
            if self.getDisplayType() == 'ascii':
                self.add_args = ''
            else:
                self.add_args = '-kb'
        if not hasattr(self, 'update_args'):
            self.update_args = ''
        if not hasattr(self, 'commit_args'):
            self.commit_args = ''

# register security information
Globals.InitializeClass(CVSFile)
